from pydantic import BaseModel, BeforeValidator, Field, PrivateAttr, model_validator
from typing_extensions import Annotated

from cg.constants import DataDelivery
from cg.models.orders.constants import OrderType
from cg.services.orders.validation.models.utils import set_null_to_false


class Order(BaseModel):
    comment: str | None = None
    customer: str = Field(min_length=1)
    delivery_type: DataDelivery
    order_type: OrderType = Field(alias="project_type")
    name: str = Field(min_length=1, max_length=64)
    skip_reception_control: Annotated[bool, BeforeValidator(set_null_to_false)] = False
    _generated_ticket_id: int | None = PrivateAttr(default=None)
    _user_id: int = PrivateAttr(default=None)

    @model_validator(mode="before")
    def convert_empty_strings_to_none(cls, data):
        if isinstance(data, dict):
            for key, value in data.items():
                if value == "":
                    data[key] = None
        return data



from cg.services.orders.validation.errors.order_errors import (
    CustomerCannotSkipReceptionControlError,
    CustomerDoesNotExistError,
    UserNotAssociatedWithCustomerError,
)
from cg.services.orders.validation.models.order import Order
from cg.store.models import User
from cg.store.store import Store


def validate_customer_exists(
    order: Order,
    store: Store,
    **kwargs,
) -> list[CustomerDoesNotExistError]:
    errors: list[CustomerDoesNotExistError] = []
    if not store.customer_exists(order.customer):
        error = CustomerDoesNotExistError()
        errors.append(error)
    return errors


def validate_user_belongs_to_customer(
    order: Order, store: Store, **kwargs
) -> list[UserNotAssociatedWithCustomerError]:
    user: User = store.get_user_by_entry_id(order._user_id)
    has_access: bool = store.is_user_associated_with_customer(
        user_id=order._user_id,
        customer_internal_id=order.customer,
    )
    errors: list[UserNotAssociatedWithCustomerError] = []
    if not (user.is_admin or has_access):
        error = UserNotAssociatedWithCustomerError()
        errors.append(error)
    return errors


def validate_customer_can_skip_reception_control(
    order: Order,
    store: Store,
    **kwargs,
) -> list[CustomerCannotSkipReceptionControlError]:
    errors: list[CustomerCannotSkipReceptionControlError] = []

    if order.skip_reception_control and not store.is_customer_trusted(order.customer):
        error = CustomerCannotSkipReceptionControlError()
        errors.append(error)
    return errors



from cg.models.analysis import AnalysisModel
from cg.models.balsamic.config import BalsamicConfigJSON
from cg.models.balsamic.metrics import BalsamicQCMetrics


class BalsamicAnalysis(AnalysisModel):
    """BALSAMIC analysis model

    Attributes:
        balsamic_config: balsamic config file attributes model
        sample_metrics: retrieved QC metrics associated to a sample
    """

    balsamic_config: BalsamicConfigJSON
    sample_metrics: dict[str, BalsamicQCMetrics]



from cg.server.dto.samples.samples_response import (
    ApplicationDTO,
    ApplicationVersionDTO,
    CustomerDto,
    SampleDTO,
)
from cg.store.models import Application, ApplicationVersion, Sample


def create_sample_dto(sample: Sample) -> SampleDTO:
    customer = create_customer_dto(sample)
    application = create_application_dto(sample.application_version.application)
    application_version = create_application_version_dto(sample.application_version)

    return SampleDTO(
        comment=sample.comment,
        control=sample.control,
        customer=customer,
        internal_id=sample.internal_id,
        name=sample.name,
        phenotype_groups=sample.phenotype_groups,
        phenotype_terms=sample.phenotype_terms,
        priority=sample.priority_human,
        reference_genome=sample.reference_genome,
        subject_id=sample.subject_id,
        is_tumour=sample.is_tumour,
        application=application,
        application_version=application_version,
        sex=sample.sex,
    )


def create_customer_dto(sample: Sample) -> CustomerDto:
    return CustomerDto(
        internal_id=sample.customer.internal_id,
        name=sample.customer.name,
    )


def create_application_dto(application: Application) -> ApplicationDTO:
    return ApplicationDTO(
        id=application.id,
        tag=application.tag,
        prep_category=application.prep_category,
        is_external=application.is_external,
        description=application.description,
        is_accredited=application.is_accredited,
        turnaround_time=application.turnaround_time,
        minimum_order=application.minimum_order,
        sequencing_depth=application.sequencing_depth,
        min_sequencing_depth=application.min_sequencing_depth,
        target_reads=application.target_reads,
        percent_reads_guaranteed=application.percent_reads_guaranteed,
        sample_amount=application.sample_amount,
        sample_volume=application.sample_volume,
        sample_concentration=application.sample_concentration,
        sample_concentration_minimum=application.sample_concentration_minimum,
        sample_concentration_maximum=application.sample_concentration_maximum,
        sample_concentration_minimum_cfdna=application.sample_concentration_minimum_cfdna,
        sample_concentration_maximum_cfdna=application.sample_concentration_maximum_cfdna,
        priority_processing=application.priority_processing,
        details=application.details,
        limitations=application.details,
        percent_kth=application.percent_kth,
        comment=application.comment,
        is_archived=application.is_archived,
        created_at=application.created_at,
        updated_at=application.updated_at,
    )


def create_application_version_dto(version: ApplicationVersion) -> ApplicationVersionDTO:
    return ApplicationVersionDTO(
        id=version.id,
        version=version.version,
        valid_from=version.valid_from,
        price_standard=version.price_standard,
        price_priority=version.price_priority,
        price_express=version.price_express,
        price_research=version.price_research,
        price_clinical_trials=version.price_clinical_trials,
        comment=version.comment,
        created_at=version.created_at,
        updated_at=version.updated_at,
        application_id=version.application_id,
    )



from pathlib import Path

from cg.constants.constants import WorkflowManager
from cg.io.json import read_json
from cg.services.analysis_starter.configurator.models.balsamic import BalsamicCaseConfig
from cg.services.analysis_starter.tracker.tracker import Tracker


class BalsamicTracker(Tracker):

    def _workflow_manager(self):
        return WorkflowManager.Slurm

    def _get_job_ids_path(self, case_id: str) -> Path:
        return Path(self.workflow_root, case_id, "analysis", "slurm_jobids.yaml")

    def _get_workflow_version(self, case_config: BalsamicCaseConfig) -> str:
        config_data: dict = read_json(case_config.sample_config)
        return config_data["analysis"]["BALSAMIC_version"]



import logging

from cg.constants.priority import Priority
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)


def case_pass_sequencing_qc(case: Case) -> bool:
    """
    Get the sequencing QC of a case. The checks are performed in the following order:
    1. If the case is a ready-made library, the ready made library QC is used.
    2. If the case is express priority, the express QC is used.
    3. If neither of the above conditions are met, it checks if all samples have enough reads.

    The order of these checks is important because the ready-made library QC is more relaxed,
    and should be checked before the express QC, and the express QC should be checked before the
    standard QC.
    """
    if is_case_ready_made_library(case):
        return ready_made_library_case_pass_sequencing_qc(case)
    if is_case_express_priority(case):
        return express_case_pass_sequencing_qc(case)
    return all(sample_has_enough_reads(sample) for sample in case.samples)


def express_case_pass_sequencing_qc(case: Case) -> bool:
    """
    Checks if all samples in an express case have enough reads.
    """
    return all(express_sample_has_enough_reads(sample) for sample in case.samples)


def express_sample_pass_sequencing_qc(sample: Sample) -> bool:
    return express_sample_has_enough_reads(sample)


def sample_pass_sequencing_qc(sample: Sample) -> bool:
    """
    Get the standard sequencing QC of a sample. The checks are performed in the following order:
    1. If the sample is express priority, the express QC is used.
    2. If the sample is a ready-made library, the ready made library QC is used.
    3. If neither of the above conditions are met, it checks if the sample has enough reads.

    The order of these checks is important because the express QC is more relaxed, and should be
    checked before the standard QC.
    """
    if is_sample_express_priority(sample):
        return express_sample_pass_sequencing_qc(sample)
    if is_sample_ready_made_library(sample):
        return ready_made_library_sample_has_enough_reads(sample)
    return sample_has_enough_reads(sample)


def ready_made_library_case_pass_sequencing_qc(case: Case) -> bool:
    """
    Checks if the sum of reads of all samples in a ready made library case is enough.

    The threshold is the expected reads for any sample in the case.
    """
    threshold: int = case.samples[0].expected_reads_for_sample
    reads: int = sum(sample.reads for sample in case.samples)
    enough_reads: bool = reads >= threshold
    if not enough_reads:
        LOG.warning(f"Case {case.internal_id} has too few reads.")
    return enough_reads


def all_samples_in_case_have_reads(case: Case) -> bool:
    """
    Check if all samples have reads.
    """
    passed_quality_check: bool = all(sample.has_reads for sample in case.samples)
    if not passed_quality_check:
        LOG.warning("Not all samples in case have reads.")
    return passed_quality_check


def any_sample_in_case_has_reads(case: Case) -> bool:
    """
    Check if any sample has reads.
    """
    passed_quality_check: bool = any(sample.has_reads for sample in case.samples)
    if not passed_quality_check:
        LOG.warning("No samples in case have reads.")
    return passed_quality_check


def is_case_express_priority(case: Case) -> bool:
    """
    Check if a case is express priority.
    """
    return case.priority == Priority.express


def express_sample_has_enough_reads(sample: Sample) -> bool:
    """
    Checks if given express sample has enough reads. Gets the threshold from the sample's
    application version.
    """
    express_reads_threshold: int = get_express_reads_threshold_for_sample(sample)
    enough_reads: bool = sample.reads >= express_reads_threshold
    if not enough_reads:
        LOG.warning(f"Sample {sample.internal_id} has too few reads.")
    return enough_reads


def get_express_reads_threshold_for_sample(sample: Sample) -> int:
    """
    Get the express reads threshold for a sample.
    """
    return round(sample.application_version.application.target_reads / 2)


def is_sample_ready_made_library(sample: Sample) -> bool:
    return sample.prep_category == SeqLibraryPrepCategory.READY_MADE_LIBRARY


def is_case_ready_made_library(case: Case) -> bool:
    """
    Check if all samples in case are ready made libraries.
    """
    return all(is_sample_ready_made_library(sample) for sample in case.samples)


def ready_made_library_sample_has_enough_reads(sample: Sample) -> bool:
    """
    Check if a given sample from a ready made library has enough reads.
    """
    if not sample.has_reads:
        LOG.warning(f"Sample {sample.internal_id} has no reads.")
    return sample.has_reads


def sample_has_enough_reads(sample: Sample) -> bool:
    """
    Check if the sample has more or equal reads than the expected reads for the sample.
    """
    enough_reads: bool = sample.reads >= sample.expected_reads_for_sample
    if not enough_reads:
        LOG.warning(f"Sample {sample.internal_id} has too few reads.")
    return enough_reads


def is_sample_express_priority(sample: Sample) -> bool:
    """
    Check if a sample is express priority.
    """
    return sample.priority == Priority.express



import logging
from functools import wraps
from pathlib import Path

from pydantic import ValidationError
from requests import ConnectionError, HTTPError
from requests.exceptions import MissingSchema

from cg.clients.freshdesk.exceptions import FreshdeskAPIException, FreshdeskModelException

LOG = logging.getLogger(__name__)


def extract_error_detail(error):
    """Extract detailed error information from HTTPError."""
    if error.response is not None:
        try:
            return error.response.json()
        except ValueError:
            return error.response.text
    return None


def log_and_raise(exception_class, message, error):
    """Log the error and raise the appropriate exception."""
    LOG.error(message)
    raise exception_class(error) from error


def handle_client_errors(func):
    """Decorator to handle and log errors in Freshdesk client methods."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPError as error:
            error_detail = extract_error_detail(error)
            log_and_raise(
                FreshdeskAPIException,
                f"Failed request to Freshdesk: {error} - Status code: "
                f"{error.response.status_code if error.response else 'N/A'}, Details: {error_detail}",
                error,
            )
        except (MissingSchema, ConnectionError) as error:
            log_and_raise(FreshdeskAPIException, f"Request to Freshdesk failed: {error}", error)
        except ValidationError as error:
            log_and_raise(
                FreshdeskModelException, f"Invalid response from Freshdesk: {error}", error
            )
        except ValueError as error:
            log_and_raise(FreshdeskAPIException, f"Operation failed: {error}", error)
        except Exception as error:
            log_and_raise(
                FreshdeskAPIException, f"Unexpected error in Freshdesk client: {error}", error
            )

    return wrapper


def prepare_attachments(attachments: list[Path]) -> list[tuple[str, tuple[str, bytes]]]:
    """Prepare the attachments for a request."""
    return [
        ("attachments[]", (attachment.name, open(attachment, "rb"))) for attachment in attachments
    ]



from cg.models.orders.constants import OrderType
from cg.services.web_services.application.models import ApplicationResponse
from cg.store.models import Application
from cg.store.store import Store


def create_application_response(app_tags: list[str]) -> ApplicationResponse:
    return ApplicationResponse(applications=app_tags)


class ApplicationsWebService:

    def __init__(self, store: Store):
        self.store = store

    def get_valid_applications(self, order_type: OrderType) -> ApplicationResponse:
        applications: list[Application] = self.store.get_active_applications_by_order_type(
            order_type
        )
        app_tags: list[str] = sorted([application.tag for application in applications])
        return create_application_response(app_tags)

    def update_application_order_types(
        self, application: Application, order_types: list[OrderType]
    ) -> None:
        self.store.delete_order_type_applications_by_application_id(application.id)
        self.store.session.add_all(
            self.store.link_order_types_to_application(
                application=application, order_types=order_types
            )
        )
        self.store.commit_to_store()



from pydantic import BaseModel, Field

from cg.constants.constants import CaseActions


class CasesRequest(BaseModel):
    action: CaseActions | None = None
    enquiry: str | None = None
    page: int = 1
    page_size: int = Field(50, alias="pageSize")



from cg.constants import Workflow

IMPLEMENTED_FASTQ_WORKFLOWS = [
    Workflow.BALSAMIC,
    Workflow.MICROSALT,
    Workflow.MIP_DNA,
    Workflow.RAREDISEASE,
    Workflow.RNAFUSION,
    Workflow.TAXPROFILER,
]

BAM_WORKFLOWS = [Workflow.NALLO]



import logging

from housekeeper.store.models import Version

from cg.apps.lims import LimsAPI
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.constants.scout import RNAFUSION_CASE_TAGS, RNAFUSION_SAMPLE_TAGS, GenomeBuild, UploadTrack
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.models.scout.scout_load_config import RnafusionLoadConfig, ScoutCancerIndividual
from cg.store.models import Analysis, CaseSample

LOG = logging.getLogger(__name__)


class RnafusionConfigBuilder(ScoutConfigBuilder):
    """Class for handling rnafusion information and files to be included in Scout upload."""

    def __init__(self, lims_api: LimsAPI):
        super().__init__(
            lims_api=lims_api,
        )
        self.case_tags: CaseTags = CaseTags(**RNAFUSION_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**RNAFUSION_SAMPLE_TAGS)

    def build_load_config(self, hk_version: Version, analysis: Analysis) -> RnafusionLoadConfig:
        """Build a rnafusion-specific load config for uploading a case to scout."""
        LOG.info("Build load config for rnafusion case")
        load_config: RnafusionLoadConfig = RnafusionLoadConfig(
            track=UploadTrack.CANCER.value,
            delivery_report=self.get_file_from_hk(
                hk_tags={HK_DELIVERY_REPORT_TAG}, hk_version=hk_version
            ),
        )
        self.add_common_info_to_load_config(load_config=load_config, analysis=analysis)
        load_config.human_genome_build = GenomeBuild.hg38
        self.include_case_files(load_config=load_config, hk_version=hk_version)
        LOG.info("Building samples")
        db_sample: CaseSample
        for db_sample in analysis.case.links:
            load_config.samples.append(
                self.build_config_sample(case_sample=db_sample, hk_version=hk_version)
            )
        return load_config

    def include_case_files(
        self, load_config: RnafusionLoadConfig, hk_version: Version = None
    ) -> None:
        """Include case level files for rnafusion case."""
        LOG.info("Including RNAFUSION specific case level files")
        for scout_key in RNAFUSION_CASE_TAGS.keys():
            self._include_file(load_config, scout_key, hk_version=hk_version)

    def include_sample_alignment_file(
        self, config_sample: ScoutCancerIndividual, hk_version: Version
    ) -> None:
        """Include the alignment file for a sample
        Try if cram file is found, if not: load bam file
        """
        sample_id: str = config_sample.sample_id
        config_sample.rna_alignment_path = self.get_sample_file(
            hk_tags=self.sample_tags.alignment_file,
            sample_id=sample_id,
            hk_version=hk_version,
        )

    def _include_file(
        self, load_config: RnafusionLoadConfig, scout_key: str, hk_version: Version
    ) -> None:
        """Include the file path associated to a scout configuration parameter if the corresponding housekeeper tags
        are found. Otherwise return None."""
        setattr(
            load_config,
            scout_key,
            self.get_file_from_hk(
                hk_tags=getattr(self.case_tags, scout_key), hk_version=hk_version
            ),
        )

    def include_sample_files(
        self, config_sample: ScoutCancerIndividual, hk_version: Version
    ) -> None:
        """Include all files that are used on sample level in Scout."""
        return None



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.taxprofiler.constants import TaxprofilerDeliveryType
from cg.services.orders.validation.order_types.taxprofiler.models.sample import TaxprofilerSample


class TaxprofilerOrder(OrderWithSamples):
    delivery_type: TaxprofilerDeliveryType
    samples: list[TaxprofilerSample]

    @property
    def enumerated_samples(self) -> enumerate[TaxprofilerSample]:
        return enumerate(self.samples)



from os.path import isfile
from pathlib import Path

from cg.services.analysis_starter.configurator.extensions.pipeline_extension import (
    PipelineExtension,
)
from cg.services.analysis_starter.configurator.file_creators.gene_panel import GenePanelFileCreator
from cg.services.analysis_starter.configurator.file_creators.managed_variants import (
    ManagedVariantsFileCreator,
)

GENE_PANEL_FILE_NAME = "gene_panels.bed"
MANAGED_VARIANTS_FILE_NAME = "managed_variants.vcf"


class RarediseaseExtension(PipelineExtension):
    """Contains Raredisease specific file creations which differ from the default Nextflow flow."""

    def __init__(
        self,
        gene_panel_file_creator: GenePanelFileCreator,
        managed_variants_file_creator: ManagedVariantsFileCreator,
    ):
        self.gene_panel_file_creator = gene_panel_file_creator
        self.managed_variants_file_creator = managed_variants_file_creator

    def configure(self, case_id: str, case_run_directory: Path) -> None:
        """Perform pipeline specific actions."""
        self.gene_panel_file_creator.create(
            case_id=case_id, file_path=_get_gene_panel_file_path(case_run_directory)
        )
        self.managed_variants_file_creator.create(
            case_id=case_id, file_path=_get_managed_variants(case_run_directory)
        )

    def do_required_files_exist(self, case_run_directory: Path) -> bool:
        return isfile(_get_gene_panel_file_path(case_run_directory)) and isfile(
            _get_managed_variants(case_run_directory)
        )


def _get_gene_panel_file_path(case_run_directory: Path) -> Path:
    return case_run_directory.joinpath(GENE_PANEL_FILE_NAME)


def _get_managed_variants(case_run_directory: Path):
    return case_run_directory.joinpath(MANAGED_VARIANTS_FILE_NAME)



import re
from collections import Counter

from cg.models.orders.sample_base import ContainerEnum, ControlEnum
from cg.services.orders.validation.constants import ALLOWED_SKIP_RC_BUFFERS, IndexEnum
from cg.services.orders.validation.errors.sample_errors import (
    BufferInvalidError,
    ConcentrationInvalidIfSkipRCError,
    ConcentrationRequiredError,
    OccupiedWellError,
    SampleError,
    SampleNameNotAvailableControlError,
    SampleNameNotAvailableError,
    WellPositionMissingError,
)
from cg.services.orders.validation.index_sequences import INDEX_SEQUENCES
from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.models.sample_aliases import IndexedSample
from cg.services.orders.validation.order_types.fastq.models.order import FastqOrder
from cg.services.orders.validation.order_types.fastq.models.sample import FastqSample
from cg.services.orders.validation.rules.utils import (
    get_application_concentration_interval,
    get_concentration_interval,
    has_sample_invalid_concentration,
    is_sample_cfdna,
)
from cg.store.models import Application
from cg.store.store import Store


class PlateSamplesValidator:

    def __init__(self, order: OrderWithSamples):
        self.wells: dict[tuple[str, str], list[int]] = {}
        self.plate_samples: list[tuple[int, Sample]] = []
        self._initialize_wells(order)

    def _initialize_wells(self, order: OrderWithSamples):
        """
        Construct a dict with keys being a (container_name, well_position) pair.
        The value will be a list of sample indices for samples located in the well.
        """
        for sample_index, sample in order.enumerated_samples:
            if sample.is_on_plate:
                self.plate_samples.append((sample_index, sample))
                key: tuple[str, str] = (sample.container_name, sample.well_position)
                if not self.wells.get(key):
                    self.wells[key] = []
                self.wells[key].append(sample_index)

    def get_occupied_well_errors(self) -> list[OccupiedWellError]:
        """Get errors for samples assigned to wells that are already occupied."""
        conflicting_samples: list[int] = []
        for samples_indices in self.wells.values():
            if len(samples_indices) > 1:
                conflicting_samples.extend(samples_indices[1:])
        return get_occupied_well_errors(conflicting_samples)

    def get_well_position_missing_errors(self) -> list[WellPositionMissingError]:
        """Get errors for samples missing well positions."""
        samples_missing_wells: list[int] = []
        for sample_index, sample in self.plate_samples:
            if not sample.well_position:
                samples_missing_wells.append(sample_index)
        return get_missing_well_errors(samples_missing_wells)


def get_occupied_well_errors(sample_indices: list[int]) -> list[OccupiedWellError]:
    return [OccupiedWellError(sample_index=sample_index) for sample_index in sample_indices]


def get_missing_well_errors(sample_indices: list[int]) -> list[WellPositionMissingError]:
    return [WellPositionMissingError(sample_index=sample_index) for sample_index in sample_indices]


def get_indices_for_repeated_sample_names(order: OrderWithSamples) -> list[int]:
    counter = Counter([sample.name for sample in order.samples])
    indices: list[int] = []
    for index, sample in order.enumerated_samples:
        if counter.get(sample.name) > 1:
            indices.append(index)
    return indices


def get_sample_name_not_available_errors(
    order: OrderWithSamples, store: Store, has_order_control: bool
) -> list[SampleError]:
    """Return errors for non-control samples with names already used in the database."""
    errors: list[SampleError] = []
    customer = store.get_customer_by_internal_id(order.customer)
    for sample_index, sample in order.enumerated_samples:
        if store.get_sample_by_customer_and_name(
            sample_name=sample.name, customer_entry_id=[customer.id]
        ):
            if is_sample_name_allowed_to_be_repeated(has_control=has_order_control, sample=sample):
                continue
            error = get_appropriate_sample_name_available_error(
                has_control=has_order_control, sample_index=sample_index
            )
            errors.append(error)
    return errors


def is_sample_name_allowed_to_be_repeated(has_control: bool, sample: Sample) -> bool:
    """
    Return whether a sample name can be used if it is already in the database.
    This is the case when the order has control samples and the sample is a control.
    """
    return has_control and sample.control in [ControlEnum.positive, ControlEnum.negative]


def get_appropriate_sample_name_available_error(
    has_control: bool, sample_index: int
) -> SampleError:
    """
    Return the appropriate error for a sample name that is not available based on whether the
    order has control samples or not.
    """
    if has_control:
        return SampleNameNotAvailableControlError(sample_index=sample_index)
    return SampleNameNotAvailableError(sample_index=sample_index)


def is_tube_container_name_redundant(sample: Sample, counter: Counter) -> bool:
    return sample.container == ContainerEnum.tube and counter.get(sample.container_name) > 1


def get_indices_for_tube_repeated_container_name(order: OrderWithSamples) -> list[int]:
    counter = Counter([sample.container_name for sample in order.samples])
    indices: list[int] = []
    for index, sample in order.enumerated_samples:
        if is_tube_container_name_redundant(sample, counter):
            indices.append(index)
    return indices


def is_invalid_well_format(sample: Sample) -> bool:
    """Check if a sample has an invalid well format."""
    correct_well_position_pattern: str = r"^[A-H]:([1-9]|1[0-2])$"
    if sample.is_on_plate:
        return not bool(re.match(correct_well_position_pattern, sample.well_position))
    return False


def is_invalid_well_format_rml(sample: IndexedSample) -> bool:
    """Check if an indexed sample has an invalid well format."""
    correct_well_position_pattern: str = r"^[A-H]:([1-9]|1[0-2])$"
    return not bool(re.match(correct_well_position_pattern, sample.well_position_rml))


def is_container_name_missing(sample: Sample) -> bool:
    """Checks if a sample is missing its container name."""
    if sample.is_on_plate and not sample.container_name:
        return True
    return False


def create_invalid_concentration_error(
    sample: FastqSample, sample_index: int, store: Store
) -> ConcentrationInvalidIfSkipRCError:
    application: Application = store.get_application_by_tag(sample.application)
    is_cfdna: bool = is_sample_cfdna(sample)
    allowed_interval: tuple[float, float] = get_application_concentration_interval(
        application=application,
        is_cfdna=is_cfdna,
    )
    return ConcentrationInvalidIfSkipRCError(
        sample_index=sample_index,
        allowed_interval=allowed_interval,
    )


def validate_concentration_interval(
    order: FastqOrder, store: Store
) -> list[ConcentrationInvalidIfSkipRCError]:
    errors: list[ConcentrationInvalidIfSkipRCError] = []
    for sample_index, sample in order.enumerated_samples:
        if application := store.get_application_by_tag(sample.application):
            allowed_interval: tuple[float, float] = get_concentration_interval(
                sample=sample, application=application
            )
            if allowed_interval and has_sample_invalid_concentration(
                sample=sample, allowed_interval=allowed_interval
            ):
                error: ConcentrationInvalidIfSkipRCError = create_invalid_concentration_error(
                    sample=sample,
                    sample_index=sample_index,
                    store=store,
                )
                errors.append(error)
    return errors


def validate_concentration_required(order: FastqOrder) -> list[ConcentrationRequiredError]:
    errors: list[ConcentrationRequiredError] = []
    for sample_index, sample in order.enumerated_samples:
        if not sample.concentration_ng_ul:
            error = ConcentrationRequiredError(sample_index=sample_index)
            errors.append(error)
    return errors


def has_multiple_applications(samples: list[IndexedSample]) -> bool:
    return len({sample.application for sample in samples}) > 1


def has_multiple_priorities(samples: list[IndexedSample]) -> bool:
    return len({sample.priority for sample in samples}) > 1


def is_index_number_missing(sample: IndexedSample) -> bool:
    """Checks if a sample is missing its index number.
    Note: Index is an attribute on the sample, not its position in the list of samples."""
    return sample.index != IndexEnum.NO_INDEX and not sample.index_number


def is_index_number_out_of_range(sample: IndexedSample) -> bool:
    """Validates that the sample's index number is in range for its specified index.
    Note: Index number is an attribute on the sample, not its position in the list of samples."""
    return sample.index_number and not (
        1 <= sample.index_number <= len(INDEX_SEQUENCES[sample.index])
    )


def validate_buffers_are_allowed(order: FastqOrder) -> list[BufferInvalidError]:
    """
    Validate that the order has only samples with buffers that allow to skip reception control.
    We can only allow skipping reception control if there is no need to exchange buffer,
    so if the sample has nuclease-free water or Tris-HCL as buffer.
    """
    errors: list[BufferInvalidError] = []
    for sample_index, sample in order.enumerated_samples:
        if sample.elution_buffer not in ALLOWED_SKIP_RC_BUFFERS:
            error = BufferInvalidError(sample_index=sample_index)
            errors.append(error)
    return errors



from abc import ABC, abstractmethod

from cg.services.delivery_message.messages.utils import (
    REMINDER_TO_DOWNLOAD_MESSAGE,
    get_caesar_delivery_path,
    get_scout_links_row_separated,
)
from cg.store.models import Case
from cg.store.store import Store


class RNAUploadMessageStrategy(ABC):
    """Abstract base class for delivery message strategies."""

    @abstractmethod
    def get_file_upload_message(self, delivery_path: str) -> str:
        """Generate the file upload message part."""
        pass


class RNAAnalysisStrategy(RNAUploadMessageStrategy):
    def get_file_upload_message(self, delivery_path: str) -> str:
        return (
            f"The analysis files are currently being uploaded to your inbox on Caesar:\n\n"
            f"{delivery_path} \n\n"
            f"{REMINDER_TO_DOWNLOAD_MESSAGE}"
        )


class RNAFastqAnalysisStrategy(RNAUploadMessageStrategy):
    def get_file_upload_message(self, delivery_path: str) -> str:
        return (
            f"The fastq and analysis files are currently being uploaded to your inbox on Caesar:\n\n"
            f"{delivery_path} \n\n"
            f"{REMINDER_TO_DOWNLOAD_MESSAGE}"
        )


class RNAFastqStrategy(RNAUploadMessageStrategy):
    def get_file_upload_message(self, delivery_path: str) -> str:
        return (
            f"The fastq files are currently being uploaded to your inbox on Caesar:\n\n"
            f"{delivery_path} \n\n"
            f"{REMINDER_TO_DOWNLOAD_MESSAGE}"
        )


class RNAScoutStrategy(RNAUploadMessageStrategy):
    def get_file_upload_message(self, delivery_path: str) -> str:
        return ""  # No file upload message needed for this case.


class RNADeliveryMessage:
    def __init__(self, store: Store, strategy: RNAUploadMessageStrategy):
        self.store = store
        self.strategy = strategy

    def create_message(self, cases: list[Case]) -> str:
        message = "Hello,\n\n"
        for case in cases:
            scout_message: str = self._get_scout_message_for_case(case=case)
            message += scout_message
        delivery_path: str = get_caesar_delivery_path(cases[0])
        file_upload_message: str = self.strategy.get_file_upload_message(delivery_path)
        return message + file_upload_message

    def _get_scout_message_for_case(self, case: Case) -> str:
        related_uploaded_dna_cases: list[Case] = self.store.get_uploaded_related_dna_cases(
            rna_case=case
        )
        scout_links: str = get_scout_links_row_separated(cases=related_uploaded_dna_cases)
        return (
            f"The analysis for case {case.name} has been uploaded to the corresponding DNA case(s) on Scout at:\n\n"
            f"{scout_links}\n\n"
        )



import logging
from pathlib import Path

from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import Workflow
from cg.constants.gene_panel import GenePanelGenomeBuild
from cg.io.txt import write_txt_with_newlines
from cg.services.analysis_starter.configurator.file_creators.nextflow.utils import get_genome_build
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class ManagedVariantsFileCreator:

    def __init__(self, scout_api: ScoutAPI, store: Store):
        self.scout_api = scout_api
        self.store = store

    def create(self, case_id: str, file_path: Path) -> None:
        content: list[str] = self._get_content(case_id)
        write_txt_with_newlines(file_path=file_path, content=content)
        LOG.info(f"Created managed variants file for case {case_id} at {file_path}")

    def _get_content(self, case_id: str) -> list[str]:
        workflow = Workflow(self.store.get_case_by_internal_id(case_id).data_analysis)
        genome_build: GenePanelGenomeBuild = get_genome_build(workflow)
        return self.scout_api.export_managed_variants(genome_build)



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_case import ExistingCase
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.nallo.constants import NalloDeliveryType
from cg.services.orders.validation.order_types.nallo.models.case import NalloCase
from cg.services.orders.validation.order_types.nallo.models.sample import NalloSample

NewCase = Annotated[NalloCase, Tag("new")]
OldCase = Annotated[ExistingCase, Tag("existing")]


class NalloOrder(OrderWithCases):
    cases: list[Annotated[NewCase | OldCase, Discriminator(has_internal_id)]]
    delivery_type: NalloDeliveryType

    @property
    def enumerated_new_samples(self) -> list[tuple[int, int, NalloSample]]:
        return [
            (case_index, sample_index, sample)
            for case_index, case in self.enumerated_new_cases
            for sample_index, sample in case.enumerated_new_samples
        ]



from enum import StrEnum

from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES

SEX_MAP = {"male": "M", "female": "F", "unknown": "unknown"}
REV_SEX_MAP = {value: key for key, value in SEX_MAP.items()}
SOURCE_TYPES = set().union(METAGENOME_SOURCES, ANALYSIS_SOURCES)


class Orderform(StrEnum):
    BALSAMIC = "1508"
    BALSAMIC_UMI = "1508"
    FASTQ = "1508"
    METAGENOME = "1508"
    FLUFFY = "1604"
    MICROSALT = "1603"
    MIP_DNA = "1508"
    MIP_RNA = "1508"
    NALLO = "long_read_nallo_analysis"
    RAREDISEASE = "1508"
    RNAFUSION = "1508"
    RML = "1604"
    SARS_COV_2 = "2184"
    MICROBIAL_FASTQ = "microbial_sequencing"
    PACBIO_LONG_READ = "pacbio_revio_sequencing"
    TAXPROFILER = "1508"
    TOMTE = "1508"

    @staticmethod
    def get_current_orderform_version(order_form: str) -> str:
        """Returns the current version of the given order form."""
        current_order_form_versions = {
            Orderform.MIP_DNA: "35",
            Orderform.RML: "20",
            Orderform.MICROSALT: "12",
            Orderform.NALLO: "1",
            Orderform.SARS_COV_2: "10",
            Orderform.MICROBIAL_FASTQ: "3",
            Orderform.PACBIO_LONG_READ: "2",
        }
        return current_order_form_versions[order_form]


REGION_CODES: dict[str, str] = {
    "Stockholm": "01",
    "Uppsala": "03",
    "Sörmland": "04",
    "Östergötland": "05",
    "Jönköpings län": "06",
    "Kronoberg": "07",
    "Kalmar län": "08",
    "Gotland": "09",
    "Blekinge": "10",
    "Skåne": "12",
    "Halland": "13",
    "Västra Götalandsregionen": "14",
    "Värmland": "17",
    "Örebro län": "18",
    "Västmanland": "19",
    "Dalarna": "20",
    "Gävleborg": "21",
    "Västernorrland": "22",
    "Jämtland Härjedalen": "23",
    "Västerbotten": "24",
    "Norrbotten": "25",
}

ORIGINAL_LAB_ADDRESSES: dict[str, str] = {
    "Unilabs Stockholm": "171 54 Solna",
    "Synlab Medilab": "183 53 Täby",
    "A05 Diagnostics": "171 65 Solna",
    "Karolinska University Hospital Solna": "171 76 Stockholm",
    "Karolinska University Hospital Huddinge": "141 86 Stockholm",
    "LaboratorieMedicinskt Centrum Gotland": "621 84 Visby",
    "Unilabs Eskilstuna Laboratorium": "631 88 Eskilstuna",
    "Norrland University Hospital": "901 85 Umeå",
    "Länssjukhuset Sundsvall": "856 43 Sundsvall",
}



"""Module to hold functions for calculations."""


def fraction_to_percent(value: float) -> float:
    """Convert a fraction to a percentage."""
    if 0.0 <= value <= 1.0:
        value *= 100
    return value


def multiply_by_million(number: float | int) -> int:
    """Multiply a given number by a million."""
    return int(number * 1_000_000)



import logging
from pathlib import Path
from typing import Hashable, Iterable

from pydantic import BaseModel, ConfigDict, constr

from cg.apps.orderform.utils import ORDER_TYPES_WITH_CASES
from cg.constants import DataDelivery
from cg.exc import OrderFormError
from cg.models.orders.constants import OrderType
from cg.models.orders.orderform_schema import OrderCase, Orderform, OrderPool
from cg.models.orders.sample_base import OrderSample
from cg.store.models import Customer

LOG = logging.getLogger(__name__)


class OrderformParser(BaseModel):
    """Class to parse orderforms"""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    samples: list[OrderSample] = []
    project_type: OrderType | None = None
    delivery_type: DataDelivery | None = None
    customer_id: constr(
        min_length=1, max_length=Customer.internal_id.property.columns[0].type.length
    ) = None
    order_comment: str | None = None
    order_name: str | None = None

    def parse_orderform(self, orderform_file: Path) -> None:
        """Parse the orderform information"""
        raise NotImplementedError

    def group_cases(self) -> dict[str, list[OrderSample]]:
        """Group samples in cases."""
        LOG.info("Group samples under respective case")
        cases = {}
        for sample in self.samples:
            case_id = sample.family_name
            if not case_id:
                continue
            if case_id not in cases:
                cases[case_id] = []
            cases[case_id].append(sample)
        if cases:
            LOG.info(f"Found cases {', '.join(cases.keys())}")
        else:
            LOG.info("Could not find any cases")
        return cases

    def get_pools(self) -> list[OrderPool]:
        """create pools from samples

        Check that all samples from one pool has the same application
        """
        pools: dict[str, OrderPool] = {}
        sample: OrderSample
        for sample in self.samples:
            pool_name = sample.pool
            if not pool_name:
                continue
            application = sample.application
            if pool_name not in pools:
                pools[pool_name] = OrderPool(
                    name=pool_name,
                    data_analysis=sample.data_analysis,
                    data_delivery=sample.data_delivery,
                    application=application,
                    samples=[sample],
                )
                continue
            # each pool must only have one application type
            pool = pools[pool_name]
            if str(pool.application) != application:
                raise OrderFormError(
                    f"different applications in pool: {pool_name} - {[pools[pool_name].application, application]}"
                )
            pool.samples.append(sample)

        return [pools[pool_name] for pool_name in pools]

    @staticmethod
    def _get_single_value(
        items_id: str, items: Iterable, attr: str, default_value: Hashable = None
    ) -> Hashable:
        """
        Get single value from a bunch of items, will raise if value not same on all items
        @param items_id:        id of items group (e.g. case-id)
        @param items:           some items (e.g. samples)
        @param attr:            one attribute of item containing a value (e.g. synopsis)
        @param default_value:   default value to use if item has an empty value of attr
        @return:                a value with the attribute of item in items
        """
        values: set[Hashable] = set(getattr(item, attr, default_value) for item in items)
        if len(values) > 1:
            raise OrderFormError(f"multiple values [{values}] for '{attr}' for '{items_id}'")

        return values.pop()

    @staticmethod
    def _get_single_set(items_id: str, items: Iterable, attr: str) -> set[Hashable]:
        """
        Get single value (set) from a bunch of items, will raise if value not same on all items
        @param items_id: id of items group (e.g. case-id)
        @param items:    some items (e.g. samples)
        @param attr:     one attribute of item containing a list (e.g. panels)
        @return:         a set with the attribute of item in items
        """
        values: set[Hashable] = set()
        for item_idx, item in enumerate(items):
            if item_idx == 0:
                values = set(getattr(item, attr)) if getattr(item, attr) else set()
            elif values != set(getattr(item, attr)) if getattr(item, attr) else set():
                raise OrderFormError(f"multiple values [{values}] for '{attr}' for '{items_id}'")
        return values

    @staticmethod
    def expand_case(case_id: str, case_samples: list[OrderSample]) -> OrderCase:
        """Fill-in information about case."""

        priority: Hashable[str] = OrderformParser._get_single_value(
            items_id=case_id, items=case_samples, attr="priority"
        )
        synopsis: Hashable[str] = OrderformParser._get_single_value(
            items_id=case_id, items=case_samples, attr="synopsis"
        )
        cohorts: set[Hashable[str]] = OrderformParser._get_single_set(
            items_id=case_id, items=case_samples, attr="cohorts"
        )
        panels: set[Hashable[str]] = OrderformParser._get_single_set(
            items_id=case_id, items=case_samples, attr="panels"
        )

        return OrderCase(
            cohorts=list(cohorts),
            name=case_id,
            samples=case_samples,
            priority=priority,
            panels=list(panels),
            synopsis=synopsis,
        )

    def generate_orderform(self) -> Orderform:
        """Generate an orderform"""
        case_objs: list[OrderCase] = []
        if self.project_type in ORDER_TYPES_WITH_CASES:
            cases_map: dict[str, list[OrderSample]] = self.group_cases()
            for case_id in cases_map:
                case_objs.append(self.expand_case(case_id=case_id, case_samples=cases_map[case_id]))
        return Orderform(
            comment=self.order_comment,
            samples=self.samples,
            cases=case_objs,
            name=self.order_name,
            customer=self.customer_id,
            delivery_type=self.delivery_type,
            project_type=self.project_type,
            pools=self.get_pools(),
        )

    def __repr__(self):
        return (
            f"OrderformParser(project_type={self.project_type},delivery_type={self.delivery_type},customer_id="
            f"{self.customer_id},order_name={self.order_name})"
        )



from cg.services.orders.validation.order_types.mip_dna.models.case import MIPDNACase
from cg.services.orders.validation.order_types.raredisease.models.case import RarediseaseCase
from cg.services.orders.validation.order_types.tomte.models.case import TomteCase

CaseContainingRelatives = MIPDNACase | RarediseaseCase | TomteCase

CaseWithSkipRC = MIPDNACase | RarediseaseCase | TomteCase



from decimal import Decimal
from json import JSONEncoder

from flask_admin import Admin
from flask_cors import CORS
from flask_wtf.csrf import CSRFProtect

from cg.apps.lims import LimsAPI
from cg.apps.tb.api import TrailblazerAPI
from cg.clients.freshdesk.freshdesk_client import FreshdeskClient
from cg.server.app_config import app_config
from cg.services.delivery_message.delivery_message_service import DeliveryMessageService
from cg.services.orders.order_service.order_service import OrderService
from cg.services.orders.order_summary_service.order_summary_service import OrderSummaryService
from cg.services.orders.storing.service_registry import (
    StoringServiceRegistry,
    setup_storing_service_registry,
)
from cg.services.orders.submitter.ticket_handler import TicketHandler
from cg.services.orders.validation.service import OrderValidationService
from cg.services.run_devices.pacbio.sequencing_runs_service import PacbioSequencingRunsService
from cg.services.sample_run_metrics_service.sample_run_metrics_service import (
    SampleRunMetricsService,
)
from cg.services.web_services.application.service import ApplicationsWebService
from cg.services.web_services.case.service import CaseWebService
from cg.services.web_services.sample.service import SampleService
from cg.store.database import initialize_database
from cg.store.store import Store
from cg.server.app_config import app_config


class FlaskLims(LimsAPI):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        config = {
            "lims": {
                "host": app_config.lims_host,
                "username": app_config.lims_username,
                "password": app_config.lims_password,
            }
        }
        super(FlaskLims, self).__init__(config)


class FlaskStore(Store):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        uri = app_config.cg_sql_database_uri
        initialize_database(uri)
        super(FlaskStore, self).__init__()


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)
        return super().default(obj)


class AnalysisClient(TrailblazerAPI):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        config = {
            "trailblazer": {
                "service_account": app_config.trailblazer_service_account,
                "service_account_auth_file": app_config.trailblazer_service_account_auth_file,
                "host": app_config.trailblazer_host,
            }
        }
        super(AnalysisClient, self).__init__(config)


cors = CORS(resources={r"/api/*": {"origins": "*"}}, supports_credentials=True)
csrf = CSRFProtect()
db = FlaskStore()

admin = Admin(name="Clinical Genomics")
lims = FlaskLims()
applications_service = ApplicationsWebService(store=db)
analysis_client = AnalysisClient()
delivery_message_service = DeliveryMessageService(store=db, trailblazer_api=analysis_client)
summary_service = OrderSummaryService(store=db, analysis_client=analysis_client)
case_service = CaseWebService(store=db)
order_service = OrderService(store=db, status_service=summary_service)
pacbio_sequencing_runs_service = PacbioSequencingRunsService(db)
sample_service = SampleService(db)
sample_run_metrics_service = SampleRunMetricsService(db)
storing_service_registry: StoringServiceRegistry = setup_storing_service_registry(
    lims=lims,
    status_db=db,
)

order_validation_service = OrderValidationService(store=db)
freshdesk_client = FreshdeskClient(
    base_url=app_config.freshdesk_url, api_key=app_config.freshdesk_api_key
)
ticket_handler = TicketHandler(
    db=db,
    client=freshdesk_client,
    system_email_id=app_config.freshdesk_order_email_id,
    env=app_config.freshdesk_environment,
)



from cg.constants import DataDelivery
from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import (
    REMINDER_TO_DOWNLOAD_MESSAGE,
    get_caesar_delivery_path,
)
from cg.store.models import Case


class TaxprofilerDeliveryMessage(DeliveryMessage):

    FILE_TYPES_BY_DATA_DELIVERY_MAP: dict[DataDelivery, str] = {
        DataDelivery.FASTQ_ANALYSIS: "fastq and analysis",
        DataDelivery.ANALYSIS_FILES: DataDelivery.ANALYSIS_FILES,
    }

    def create_message(self, cases: list[Case]) -> str:
        delivery_path: str = get_caesar_delivery_path(cases[0])
        file_types: str = self.FILE_TYPES_BY_DATA_DELIVERY_MAP[cases[0].data_delivery]
        return (
            f"Hello,\n\n"
            f"The {file_types} files for this order are currently being uploaded to your inbox on Caesar at:\n\n"
            f"{delivery_path} \n\n"
            f"{REMINDER_TO_DOWNLOAD_MESSAGE}"
        )



from http import HTTPStatus

from flask import Blueprint, abort, g, jsonify, request

from cg.server.endpoints.utils import before_request
from cg.server.ext import db
from cg.store.models import Customer, Pool

POOLS_BLUEPRINT = Blueprint("pools", __name__, url_prefix="/api/v1")
POOLS_BLUEPRINT.before_request(before_request)


@POOLS_BLUEPRINT.route("/pools")
def parse_pools():
    """Return pools."""
    customers: list[Customer] | None = (
        g.current_user.customers if not g.current_user.is_admin else None
    )
    pools: list[Pool] = db.get_pools_to_render(
        customers=customers, enquiry=request.args.get("enquiry")
    )
    parsed_pools: list[dict] = [pool_obj.to_dict() for pool_obj in pools[:30]]
    return jsonify(pools=parsed_pools, total=len(pools))


@POOLS_BLUEPRINT.route("/pools/<pool_id>")
def parse_pool(pool_id):
    """Return a single pool."""
    pool: Pool = db.get_pool_by_entry_id(entry_id=pool_id)
    if pool is None:
        return abort(HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (pool.customer not in g.current_user.customers):
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**pool.to_dict())



from typing import Annotated, Any
from pydantic import BaseModel, BeforeValidator, Field, ValidationError, ConfigDict
from cg.store.models import Sample


# Validator
def str_to_bool(value: str) -> bool:
    if value == "TRUE":
        return True
    elif value == "FALSE":
        return False
    raise ValidationError(f"String {value} cannot be turned to bool.")


# Models
class ParsedSampleResults(BaseModel):
    sample_name: str = Field(alias="Sample")
    selection: str = Field(alias="Selection")
    region_code: str = Field(alias="Region Code")
    ticket: int = Field(alias="Ticket")
    pct_n_bases: float = Field(alias="%N_bases")
    pct_10x_coverage: float = Field(alias="%10X_coverage")
    passes_qc: Annotated[bool, BeforeValidator(str_to_bool)] = Field(alias="QC_pass")
    lineage: str = Field(alias="Lineage")
    pangolin_data_version: str = Field(alias="Pangolin_data_version")
    voc: str = Field(alias="VOC")
    mutations: str = Field(alias="Mutations")


class MutantPoolSamples(BaseModel):
    samples: list[Sample]
    external_negative_control: Sample
    internal_negative_control: Sample

    model_config = ConfigDict(arbitrary_types_allowed=True)


class SamplePoolAndResults(BaseModel):
    pool: MutantPoolSamples
    results: dict[str, ParsedSampleResults]


class SampleQualityResults(BaseModel):
    sample_id: str
    passes_qc: bool
    passes_reads_threshold: bool
    passes_mutant_qc: bool | None = None


class SamplesQualityResults(BaseModel):
    internal_negative_control: SampleQualityResults
    external_negative_control: SampleQualityResults
    samples: list[SampleQualityResults]

    @property
    def total_samples_count(self) -> int:
        return len(self.samples)

    @property
    def passed_samples_count(self) -> int:
        samples_pass_qc: list[bool] = [sample_result.passes_qc for sample_result in self.samples]
        return sum(samples_pass_qc)

    @property
    def failed_samples_count(self) -> int:
        return self.total_samples_count - self.passed_samples_count


class CaseQualityResult(BaseModel):
    passes_qc: bool
    internal_negative_control_passes_qc: bool
    external_negative_control_passes_qc: bool
    fraction_samples_passes_qc: bool


class MutantQualityResult(BaseModel):
    case_quality_result: CaseQualityResult
    samples_quality_results: SamplesQualityResults
    summary: str

    @property
    def passes_qc(self) -> bool:
        return self.case_quality_result.passes_qc


class MutantReport(BaseModel):
    summary: str
    case: dict[str, Any]
    samples: dict[str, Any]



from abc import ABC, abstractmethod


class RunNamesService(ABC):

    def __init__(self, run_directory: str):
        self.run_directory = run_directory

    @abstractmethod
    def get_run_names(self) -> list[str]:
        """Get all the run names from a run directory."""
        pass



import logging
from pathlib import Path

from cg.apps.environ import environ_email
from cg.exc import MissingConfigFilesError
from cg.meta.workflow.fastq import MipFastqHandler
from cg.models.cg_config import MipConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.configurator.file_creators.gene_panel import GenePanelFileCreator
from cg.services.analysis_starter.configurator.file_creators.managed_variants import (
    ManagedVariantsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.mip_dna_config import (
    MIPDNAConfigFileCreator,
)
from cg.services.analysis_starter.configurator.models.mip_dna import MIPDNACaseConfig
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)

MIP_DNA_CONFIG_FILE_NAME = "pedigree.yaml"
MIP_DNA_GENE_PANEL_FILE_NAME = "gene_panels.bed"
MIP_DNA_MANAGED_VARIANTS_FILE_NAME = "managed_variants.vcf"


class MIPDNAConfigurator(Configurator):
    def __init__(
        self,
        cg_mip_config: MipConfig,
        config_file_creator: MIPDNAConfigFileCreator,
        fastq_handler: MipFastqHandler,
        gene_panel_file_creator: GenePanelFileCreator,
        managed_variants_file_creator: ManagedVariantsFileCreator,
        store: Store,
    ):
        self.conda_binary = cg_mip_config.conda_binary
        self.conda_environment = cg_mip_config.conda_env
        self.pipeline_binary = cg_mip_config.script
        self.pipeline_command = cg_mip_config.workflow
        self.pipeline_config_path = cg_mip_config.mip_config
        self.root = cg_mip_config.root

        self.config_file_creator = config_file_creator
        self.fastq_handler = fastq_handler
        self.gene_panel_file_creator = gene_panel_file_creator
        self.managed_variants_file_creator = managed_variants_file_creator
        self.store = store

    def configure(self, case_id: str, **flags) -> MIPDNACaseConfig:
        LOG.info(f"Configuring case {case_id}")
        self._create_run_directory(case_id)
        self.fastq_handler.link_fastq_files(case_id)
        self.config_file_creator.create(
            case_id=case_id,
            bed_flag=flags.get("panel_bed"),
            file_path=self._get_config_file_path(case_id=case_id),
        )
        self.gene_panel_file_creator.create(
            case_id=case_id, file_path=self._get_gene_panel_file_path(case_id)
        )
        self.managed_variants_file_creator.create(
            case_id=case_id, file_path=self._get_managed_variants_file_path(case_id)
        )
        config: MIPDNACaseConfig = self.get_config(case_id=case_id, **flags)
        LOG.info(f"Case {case_id} configured successfully")
        return config

    def get_config(self, case_id: str, **flags) -> MIPDNACaseConfig:
        case: Case = self.store.get_case_by_internal_id_strict(case_id)
        config = MIPDNACaseConfig(
            case_id=case_id,
            conda_binary=self.conda_binary,
            conda_environment=self.conda_environment,
            pipeline_binary=self.pipeline_binary,
            pipeline_command=self.pipeline_command,
            pipeline_config_path=self.pipeline_config_path,
            email=environ_email(),
            slurm_qos=case.slurm_priority,
            use_bwa_mem=flags.get("use_bwa_mem", False),
        )
        config = self._set_flags(config=config, **flags)
        self._ensure_required_config_files_exist(case_id)
        return config

    def _get_gene_panel_file_path(self, case_id: str) -> Path:
        return Path(self._get_run_directory(case_id), MIP_DNA_GENE_PANEL_FILE_NAME)

    def _get_managed_variants_file_path(self, case_id: str) -> Path:
        return Path(self._get_run_directory(case_id), MIP_DNA_MANAGED_VARIANTS_FILE_NAME)

    def _get_config_file_path(self, case_id: str) -> Path:
        return Path(self._get_run_directory(case_id), MIP_DNA_CONFIG_FILE_NAME)

    def _get_run_directory(self, case_id: str) -> Path:
        return Path(self.root, case_id)

    def _create_run_directory(self, case_id: str) -> Path:
        path: Path = self._get_run_directory(case_id)
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _ensure_required_config_files_exist(self, case_id: str) -> None:
        """Ensures that the case run directory has a MIP config file, a gene panel file and a
        managed variants file.
        Raises:
            MissingConfigFilesError if any of those files are not present.
        """
        config_file_path: Path = self._get_config_file_path(case_id)
        gene_panel_file_path: Path = self._get_gene_panel_file_path(case_id)
        managed_variants_file_path: Path = self._get_managed_variants_file_path(case_id)
        if not (
            config_file_path.exists()
            and gene_panel_file_path.exists()
            and managed_variants_file_path.exists()
        ):
            raise MissingConfigFilesError(
                "Ensure config file, gene panel and managed variants files exist."
            )



import logging
import rich_click as click

from cg.models.cg_config import CGConfig
from cg.services.sequencing_qc_service.sequencing_qc_service import SequencingQCService

LOG = logging.getLogger(__name__)


@click.command(name="sequencing-qc", help="Perform sequencing QC")
@click.pass_obj
def sequencing_qc(context: CGConfig):
    service: SequencingQCService = context.sequencing_qc_service
    service.run_sequencing_qc()



from cg.constants.constants import SampleStatus
from cg.exc import AuthorisationError
from cg.server.dto.samples.requests import CollaboratorSamplesRequest, SamplesRequest
from cg.server.dto.samples.samples_response import SamplesResponse
from cg.services.web_services.sample.utils import (
    create_samples_response,
    get_cancel_comment,
    get_confirmation_message,
    get_start_and_finish_indexes_from_request,
)
from cg.store.models import Customer, Sample, User
from cg.store.store import Store


class SampleService:
    def __init__(self, store: Store):
        self.store = store

    def cancel_sample(self, sample_id: int, user_email: str) -> None:
        self.store.mark_sample_as_cancelled(sample_id)
        self.store.decouple_sample_from_cases(sample_id)
        self._add_cancel_comment(sample_id=sample_id, user_email=user_email)

    def _add_cancel_comment(self, sample_id: int, user_email: str) -> None:
        if user := self.store.get_user_by_email(user_email):
            comment = get_cancel_comment(user.name)
            self.store.update_sample_comment(sample_id=sample_id, comment=comment)

    def cancel_samples(self, sample_ids: list[int], user_email: str) -> str:
        case_ids = self.store.get_case_ids_for_samples(sample_ids)

        for sample_id in sample_ids:
            self.cancel_sample(sample_id=sample_id, user_email=user_email)

        self.store.delete_cases_without_samples(case_ids)
        remaining_cases = self.store.filter_cases_with_samples(case_ids)
        return get_confirmation_message(sample_ids=sample_ids, case_ids=remaining_cases)

    def get_collaborator_samples(self, request: CollaboratorSamplesRequest) -> SamplesResponse:
        samples: list[Sample] = self.store.get_collaborator_samples(request)
        return create_samples_response(samples)

    def get_samples(self, request: SamplesRequest, user: User) -> tuple[list[dict], int]:
        if request.status in SampleStatus.statuses():
            if not user.is_admin:
                raise AuthorisationError()
            else:
                return self._get_samples_handled_by_status(request=request)
        customers: list[Customer] | None = None if user.is_admin else user.customers
        samples, total = self.store.get_samples_by_customers_and_pattern(
            pattern=request.enquiry,
            customers=customers,
            limit=request.page_size,
            offset=(request.page - 1) * request.page_size,
        )
        parsed_samples: list[dict] = [sample.to_dict() for sample in samples]
        return parsed_samples, total

    def _get_samples_handled_by_status(self, request: SamplesRequest) -> tuple[list[dict], int]:
        """Get samples based on the provided status."""
        if request.status == SampleStatus.INCOMING:
            samples: list[Sample] = self.store.get_samples_to_receive()
        elif request.status == SampleStatus.LABPREP:
            samples: list[Sample] = self.store.get_samples_to_prepare()
        else:
            samples: list[Sample] = self.store.get_samples_to_sequence()
        start, finish = get_start_and_finish_indexes_from_request(request)
        parsed_samples: list[dict] = [sample.to_dict() for sample in samples[start:finish]]
        return parsed_samples, len(parsed_samples)



"""Observations API."""

import logging
from pathlib import Path

from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants.constants import CustomerId
from cg.constants.observations import LoqusdbInstance
from cg.constants.sample_sources import SourceType
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.exc import AnalysisNotCompletedError, LoqusdbUploadCaseError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.cg_config import CGConfig, CommonAppConfig
from cg.models.observations.input_files import (
    BalsamicObservationsInputFiles,
    MipDNAObservationsInputFiles,
    NalloObservationsInputFiles,
    RarediseaseObservationsInputFiles,
)
from cg.store.models import Analysis, Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class ObservationsAPI:
    """API to manage Loqusdb observations."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        self.store: Store = config.status_db
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.analysis_api: AnalysisAPI = analysis_api
        self.loqusdb_config: CommonAppConfig = config.loqusdb
        self.loqusdb_rd_lwp_config: CommonAppConfig = config.loqusdb_rd_lwp
        self.loqusdb_wes_config: CommonAppConfig = config.loqusdb_wes
        self.loqusdb_somatic_config: CommonAppConfig = config.loqusdb_somatic
        self.loqusdb_tumor_config: CommonAppConfig = config.loqusdb_tumor
        self.loqusdb_somatic_lymphoid_config: CommonAppConfig = config.loqusdb_somatic_lymphoid
        self.loqusdb_somatic_myeloid_config: CommonAppConfig = config.loqusdb_somatic_myeloid
        self.loqusdb_somatic_exome_config: CommonAppConfig = config.loqusdb_somatic_exome

    def upload(self, case_id: str) -> None:
        """
        Upload observations to Loqusdb.

        Raises:
            LoqusdbUploadCaseError: If case is not eligible for Loqusdb uploads
        """
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        if self.is_case_eligible_for_observations_upload(case):
            self.load_observations(case=case)
        else:
            LOG.error(f"Case {case.internal_id} is not eligible for observations upload")
            raise LoqusdbUploadCaseError

    def get_observations_input_files(
        self, case: Case
    ) -> (
        BalsamicObservationsInputFiles
        | MipDNAObservationsInputFiles
        | NalloObservationsInputFiles
        | RarediseaseObservationsInputFiles
    ):
        """Return input files from a case to upload to Loqusdb."""
        analysis: Analysis | None = case.latest_completed_analysis
        if not analysis:
            raise AnalysisNotCompletedError(f"Case {case.internal_id} has no completed analyses")
        hk_version: Version = self.housekeeper_api.get_version_by_id(
            analysis.housekeeper_version_id
        )
        return self.get_observations_files_from_hk(hk_version=hk_version, case_id=case.internal_id)

    def get_loqusdb_api(self, loqusdb_instance: LoqusdbInstance) -> LoqusdbAPI:
        """Returns a Loqusdb API for the given Loqusdb instance."""
        loqusdb_config_map: dict = {
            LoqusdbInstance.LWP: self.loqusdb_rd_lwp_config,
            LoqusdbInstance.WGS: self.loqusdb_config,
            LoqusdbInstance.WES: self.loqusdb_wes_config,
            LoqusdbInstance.SOMATIC: self.loqusdb_somatic_config,
            LoqusdbInstance.TUMOR: self.loqusdb_tumor_config,
            LoqusdbInstance.SOMATIC_LYMPHOID: self.loqusdb_somatic_lymphoid_config,
            LoqusdbInstance.SOMATIC_MYELOID: self.loqusdb_somatic_myeloid_config,
            LoqusdbInstance.SOMATIC_EXOME: self.loqusdb_somatic_exome_config,
        }
        loqusdb_config = loqusdb_config_map[loqusdb_instance]
        return LoqusdbAPI(
            binary_path=loqusdb_config.binary_path,
            config_path=loqusdb_config.config_path,
        )

    @staticmethod
    def is_duplicate(
        case: Case,
        loqusdb_api: LoqusdbAPI,
        profile_vcf_path: Path | None = None,
        profile_threshold: float | None = None,
    ) -> bool:
        """Check if a case has already been uploaded to Loqusdb."""
        loqusdb_case: dict = loqusdb_api.get_case(case_id=case.internal_id)
        duplicate = (
            loqusdb_api.get_duplicate(
                profile_vcf_path=profile_vcf_path, profile_threshold=profile_threshold
            )
            if profile_vcf_path and profile_threshold
            else None
        )
        return bool(loqusdb_case or duplicate or case.loqusdb_uploaded_samples)

    def update_statusdb_loqusdb_id(self, samples: list[Sample], loqusdb_id: str | None) -> None:
        """Update Loqusdb ID field in StatusDB for each of the provided samples."""
        for sample in samples:
            sample.loqusdb_id = loqusdb_id
        self.store.commit_to_store()

    def is_customer_eligible_for_observations_upload(self, customer_id: str) -> bool:
        """Return whether the customer has been whitelisted for uploading observations."""
        if customer_id not in self.loqusdb_customers:
            LOG.error(f"Customer {customer_id} is not whitelisted for Loqusdb uploads")
            return False
        return True

    def is_sequencing_method_eligible_for_observations_upload(self, case_id: str) -> bool:
        """Return whether a sequencing method is valid for observations upload."""
        sequencing_method: SeqLibraryPrepCategory | None = self.analysis_api.get_data_analysis_type(
            case_id
        )
        if sequencing_method not in self.loqusdb_sequencing_methods:
            LOG.error(f"Sequencing method {sequencing_method} is not supported by Loqusdb uploads")
            return False
        return True

    def is_sample_source_type_ffpe(self, case_id: str) -> bool:
        source_type: str | None = self.analysis_api.get_case_source_type(case_id)
        if source_type and SourceType.FFPE.lower() not in source_type.lower():
            return True
        LOG.error(f"Source type {source_type} is not supported for Loqusdb uploads")
        return False

    @property
    def loqusdb_customers(self) -> list[CustomerId]:
        """Customers that are eligible for Loqusdb uploads."""
        raise NotImplementedError

    @property
    def loqusdb_sequencing_methods(self) -> list[str]:
        """Sequencing methods that are eligible for Loqusdb uploads."""
        raise NotImplementedError

    def load_observations(self, case: Case) -> None:
        """Load observation counts to Loqusdb."""
        raise NotImplementedError

    def is_case_eligible_for_observations_upload(self, case: Case) -> bool:
        """Return whether a case is eligible for observations upload."""
        raise NotImplementedError

    def get_observations_files_from_hk(
        self, hk_version: Version, case_id: str
    ) -> MipDNAObservationsInputFiles | BalsamicObservationsInputFiles:
        """Return observations files given a Housekeeper version."""
        raise NotImplementedError

    def delete_case(self, case: Case) -> None:
        """Delete case observations from Loqusdb."""
        raise NotImplementedError
